/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2008-2011 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package com.sun.pkg.bootstrap;

import com.sun.pkg.client.ImagePlanProgressTracker;
import com.sun.pkg.client.Image;
import com.sun.pkg.client.Fmri;
import com.sun.pkg.client.SystemInfo;
import java.io.*;
import java.net.Proxy;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.text.MessageFormat;
import java.util.Properties;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.Vector;
import java.util.logging.ConsoleHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

/*
 * Limitations of the PkgBootstrap IPS client
 * 
 * - doesn't handle mirrors of authorities (repositories)
 * - doesn't handle multiple repositories - always uses the preferred_repository
 * - doesn't handle filters
 * - will not retrieve catalogs from repository - catalogs must be cached
 */
import java.util.logging.SimpleFormatter;

public class Bootstrap 
{
    public static final String WIN_CONFIG_TOOL = "updatetoolconfig.bat";
    public static final String UNIX_CONFIG_TOOL = "updatetoolconfig";
    public static final String WIN_NOTIFIER_APP = "Update Tool Notifier.exe";
    public static final String UNIX_NOTIFIER_APP = "updatetool";
    private static final String eol = System.getProperty("line.separator");

    private static ImagePlanProgressTracker tracker = null;

    static Properties defaultProperties = new Properties();
    static {
        defaultProperties.setProperty("image.path", ".");
        defaultProperties.setProperty("install.pkg", "true");
        defaultProperties.setProperty("install.updatetool", "false");
        defaultProperties.setProperty("proxy.URL", "");
        defaultProperties.setProperty("optin.update.notification", "true");
        defaultProperties.setProperty("uuid", "");
    }
    
    static class BootstrapFormatter extends SimpleFormatter
    {
        @Override
        public String format(LogRecord record)
        {
            return formatMessage(record) + eol;
        }
    }

    static public void main(String argv[])
    {
        ResourceBundle rb = ResourceBundle.getBundle("com/sun/pkg/bootstrap/messages");
        try {
            Properties props = new Properties(defaultProperties);
            if (argv.length > 0) {
                String propfilename = argv[0];
                FileInputStream fis = new FileInputStream(propfilename);
                props.load(fis);
                fis.close();
            }
            _main(props);
        }
        catch (UnknownHostException e) {
            System.err.println(MessageFormat.format(rb.getString("unknownhost"),
                    e.getMessage()));
            System.exit(4);
        }
        catch (FileNotFoundException e) {
            System.err.println(MessageFormat.format(rb.getString("cannotfindfile"),
                    e.getMessage()));
            System.exit(2);
        }
        catch (IOException e) {
            System.err.println(MessageFormat.format(rb.getString("ioexception"),
                    e.getMessage()));
            System.exit(3);
        }
        catch (Exception e) {
            System.err.println(MessageFormat.format(rb.getString("unrecognized"),
                    e.getMessage()));
            e.printStackTrace();
            System.exit(1);
        }
        System.exit(0);
    }
    
    static public void main(Properties mainprops, Logger log) throws Exception
    {
        Properties props = new Properties(defaultProperties);
        props.putAll(mainprops);
        _main(props, log);
    }
    
    static void _main(Properties props) throws Exception
    {
        // System properties (passed via -D on command line) override
        // all other properties. Load them.
        loadSystemProperties(props);

        Handler h = new ConsoleHandler();
        h.setFormatter(new BootstrapFormatter());
        
        // redirect log messages from pkg Java API to console
        Logger cl = Logger.getLogger("com.sun.pkg.client");
        cl.setUseParentHandlers(false);
        cl.addHandler(h);

        Logger l = Logger.getLogger("com.sun.pkg.bootstrap.Bootstrap", "com/sun/pkg/bootstrap/messages");
        l.setUseParentHandlers(false);
        l.addHandler(h);
        _main(props, l);
    }
    
    static void _main(Properties props, Logger log) throws Exception
    {
        String imagedir = props.getProperty("image.path").trim();
        String pkg_list[] = { "pkg-toolkit-incorporation", "pkg" };
        String updatetool_list[] = { "updatetool" };

        String proxyStr = props.getProperty("proxy.URL").trim();
        if (proxyStr.length() > 0) {
            try {
                URL proxy = new URL(proxyStr);
                URLConnection uc = proxy.openConnection();
                uc.setDoOutput(true);
                uc.connect();
                uc.getOutputStream().close();
            }
            catch (IOException ioe) {
                log.log(Level.SEVERE, "badproxy", proxyStr);
                throw ioe;
            }
            catch (IllegalArgumentException iae) {
                log.log(Level.SEVERE, "badproxy", proxyStr);
                throw new IOException("invalid proxy: " + iae.getMessage());
            }
            catch (NullPointerException npe) {
                log.log(Level.SEVERE, "badproxy", proxyStr);
                throw new IOException("invalid proxy: " + proxyStr);
            }
        }
        SystemInfo.initUpdateToolProps(props);
        Image img = new Image(imagedir);
        img.setClientName("pkg-bootstrap");
        Proxy p = SystemInfo.getProxy();
        if (p == null) {
            if (Boolean.parseBoolean(System.getProperty(
                                            "java.net.useSystemProxies"))) {
                log.log(Level.INFO, "proxysys");
            } else {
                log.log(Level.INFO, "proxydirect");
            }
        } else {
            log.log(Level.INFO, "proxy", p);
        }

        /*
         * Only set or override uuids on publishers when the caller
         * has supplied a uuid value. If no uuid is supplied by the caller,
         * then existing uuids, if any exist, will be left intact.
         */
        String uuid = props.getProperty("uuid").trim();
        if (uuid.length() > 0) {
            for (String a : img.getAuthorityNames()) {
                img.setAuthority(a, null, uuid);
            }
        }
        img.setProperty("send-uuid", "True");

        img.saveConfig();
        img.hideMetaDirectory();
        img.refreshCatalog(img.getPreferredAuthorityName());

        log.log(Level.INFO, "imagepath", img.getRootDirectory().getCanonicalPath());

        // If the caller has not set a tracker, then set our CLI oriented 
        // tracker.
        if (getTracker() == null) {
            setTracker(new BootstrapProgressTracker(log));
        }

        if (Boolean.parseBoolean(props.getProperty("install.pkg").trim())) {

            // Trim the pkg_list to only those packages that are not already
            // installed. We mainly do this to avoid upgrading an
            // already installed version of pkg-toolkit-incorporation
            // (see bug 2098)
            String[] install_pkg_list = trimPackageList(img, pkg_list);

            if (install_pkg_list != null && install_pkg_list.length > 0) {
                log.log(Level.INFO, "installingpkg");
                img.installPackages(getTracker(), true, true, install_pkg_list);
            }
        }
        
        if (Boolean.parseBoolean(props.getProperty("install.updatetool").trim())) {
            log.log(Level.INFO, "installingupdatetool");
            img.installPackages(getTracker(), true, true, updatetool_list);
        }

        /* 
         * Register and start the notifier only if we have installed updatetool
         * and the user has opt'ed in and we are not root.  See issue 729
         * for details.
         */
        if (Boolean.parseBoolean(
                      props.getProperty("install.updatetool").trim()) &&
            Boolean.parseBoolean(
                      props.getProperty("optin.update.notification").trim()) &&
            ! System.getProperty("user.name").equals("root")) {
            int status = Bootstrap.registerNotifier(props, log);
            // A status of 1 indicates the notifier is already registered.
            // See issue 1917
            if (status != 1) {
                String args;

                if (status == 0)
                    args = new String("--check-and-exit");
                else
                    args = new String("");

                log.log(Level.INFO, "startingnotifier");
                Bootstrap.startNotifier(props, args, log);
            }
        }

        log.log(Level.INFO, "completemsg");
    }

    static public void loadSystemProperties(Properties props) {
        // If any of our properties is set as a system property then
        // load it into our properties object. We use defaultProperties
        // as the list of properties we know about.
        Enumeration keys = defaultProperties.keys();
        while (keys.hasMoreElements()) {
            String name = (String)keys.nextElement();
            String value = System.getProperty(name);
            if (value != null) {
                //System.out.println("System property: " + name + "=" + value);
                props.setProperty(name, value);
            }
        }
    }

    /**
     * Trims a list of packages to only those that are not installed
     * in the image.
     *
     * @param img
     *      The install image to examine
     *
     * @param pkg_list
     *      List of package names to trim
     *
     * @return The package names in pkg_list that are not installed
     *
     * @throws java.io.Exception
     */
    static public String[] trimPackageList(Image img, String[] pkg_list)
            throws IOException {
            Vector<String> new_pkg_list = new Vector();

            // Scan the pkg_list. If a package is already installed then
            // don't install it.
            for (String pkg_name : pkg_list) {
                Fmri fmri = img.getVersionInstalled(pkg_name);
                if (fmri == null) {
                    // Package not installed, we want to install it
                    new_pkg_list.add(pkg_name);
                }
            }

            return new_pkg_list.toArray(new String[0]);
    }

    public static void startNotifier(Properties props, String args, Logger log) throws Exception
    {
        String imagedir = props.getProperty("image.path").trim();
        String execApp[] = new String[4] ;

        // Determine which tool to start:
        if (isWindows())
            execApp[0] = new String(imagedir + File.separator + 
                                "updatetool" + File.separator + "lib" + 
                                File.separator + WIN_NOTIFIER_APP);
        else
            execApp[0] = new String(imagedir + File.separator + 
                                "updatetool" + File.separator + "bin" + 
                                File.separator + UNIX_NOTIFIER_APP);

        execApp[1] = new String("--notifier");  
        execApp[2] = new String("--silentstart");  
        execApp[3] = args;

        try {
            Runtime rt = Runtime.getRuntime();
            Process proc = rt.exec(execApp);
        } catch (Throwable t) {
            log.log(Level.INFO, "startingnotifierexception", t.getLocalizedMessage());
        }
    }

    public static int registerNotifier(Properties props, Logger log) throws Exception
    {
        String imagedir = props.getProperty("image.path").trim();
        String execApp[] = new String[2] ;

        // Determine which tool to use to register:
        if (isWindows())
            execApp[0] = new String(imagedir + File.separator + 
                                "updatetool" + File.separator + "bin" + 
                                File.separator + WIN_CONFIG_TOOL);
        else
            execApp[0] = new String(imagedir + File.separator + 
                                "updatetool" + File.separator + "bin" + 
                                File.separator + UNIX_CONFIG_TOOL);

        execApp[1] = new String("--register"); 

        // Verify the config tool exists.  If it does not then we
        // can not register the notifier.
        try {
            File file = new File(execApp[0]);
 
            if (!file.exists())
            {
                log.log(Level.INFO, "registeringnotifierexit3");
                return 2; 
            }
        }
        catch (SecurityException e) {
            log.log(Level.INFO, "registeringnotifierexception", e.getLocalizedMessage());
            return 2; 
        }

        try {

            Runtime rt = Runtime.getRuntime();
            Process proc = rt.exec(execApp);

            // Clear the output streams... (stdin)
            InputStream stdin = proc.getInputStream();
            InputStreamReader isr = new InputStreamReader(stdin);
            BufferedReader br = new BufferedReader(isr);
            String line = null;
            while ( (line = br.readLine()) != null) { } // We ignore the output. 

            // Clear the output streams... (stderr)
            InputStream stderr = proc.getErrorStream();
            isr = new InputStreamReader(stderr);
            br = new BufferedReader(isr);
            while ( (line = br.readLine()) != null) { } // We ignore the output.

            int exitVal = proc.waitFor();

            if (exitVal == 0)
                log.log(Level.INFO, "registeringnotifierexit0");
            else if (exitVal == 1)
                log.log(Level.INFO, "registeringnotifierexit1");
            else
                log.log(Level.INFO, "registeringnotifierexit2", exitVal);

            return exitVal;

        } catch (Throwable t) {
            log.log(Level.INFO, "registeringnotifierexception", t.getLocalizedMessage());
            return 2;
        }
    }

    public static void setTracker(ImagePlanProgressTracker tracker) {
        Bootstrap.tracker = tracker;
    }

    public static ImagePlanProgressTracker getTracker() {
        return Bootstrap.tracker;
    }

    static private boolean isWindows() {
         String osname = System.getProperty("os.name");
         return osname.startsWith("Windows");
    }
}
